use crate::db::{db_new_user, db_user_login};
use crate::jwt::token_factory;
use crate::state::CaptchaState;
use crate::ServerState;

use actix_web::web::Data;
use actix_web::{post, web::Json, HttpResponse, Responder, Result};
use argon2::password_hash::rand_core::RngCore;
use argon2::password_hash::*;
use biosvg::BiosvgBuilder;
use log::*;
use serde::{Deserialize, Serialize};

#[derive(Debug, Serialize, Deserialize)]
pub struct LoginData {
    username: String,
    password: String,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct LoginResponse {
    username: String,
    token: String,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct RegisterData {
    username: String,
    password: String,
    captcha: String,
}

#[post("/register")]
pub async fn register(
    data: Json<RegisterData>,
    state: Data<ServerState>,
) -> Result<impl Responder> {
    db_new_user(data.username.clone(), data.password.clone(), &state.pool).await;
    info!("User: {} registered", data.username);
    Ok(HttpResponse::Ok().json("User registered"))
}

#[post("/login")]
pub async fn login(data: Json<LoginData>, state: Data<ServerState>) -> Result<impl Responder> {
    let result = db_user_login(data.username.clone(), data.password.clone(), &state.pool).await;

    match result {
        Some(_) => {
            return Ok(HttpResponse::Ok().json(LoginResponse {
                username: data.username.clone(),
                token: token_factory(&data.username).unwrap(),
            }));
        }
        None => {
            info!("User \"{}\" failed to log in", data.username);
            return Ok(HttpResponse::BadRequest().json("Error"));
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct CaptchaResponse {
    captcha_svg: String,
    captcha_id: i32,
}

/// Request a captcha from the captcha service
#[post("/captcha")]
pub async fn captcha_request(cstate: Data<CaptchaState>) -> Result<impl Responder> {
    // This might block the thread a bit too long
    let (answer, svg) = get_captcha();

    let id = rand_core::OsRng.next_u32() as i32;

    let cresponse = CaptchaResponse {
        captcha_svg: svg.clone(),
        captcha_id: id,
    };

    // This is bad in about every way i can think of
    // It might just be better to hit the database every time, and let the database
    // handle rng and maybe set a trigger to delete old captchas
    match cstate.capthca_db.lock() {
        Ok(mut db) => {
            if (db.len() as i32) > 100 {
                // To prevent the database from growing too large
                // Replace with a proper LRU cache or circular buffer
                db.remove(&(id % 100)); // This is terrible
            }
            db.insert(id, answer.clone()); // We do not care about collisions
            return Ok(HttpResponse::Ok().json(cresponse));
        }
        Err(_) => {
            // This shouldnt happen
            error!("Failed to lock captcha database");
            return Ok(HttpResponse::InternalServerError().json("Error"));
        }
    }
}

/// Returns a new captcha in the form of a tuple (answer, svg)
fn get_captcha() -> (String, String) {
    BiosvgBuilder::new()
        .length(4)
        .difficulty(6)
        .colors(vec![
            // Feel free to change these
            "#0078D6".to_string(),
            "#aa3333".to_string(),
            "#f08012".to_string(),
            "#33aa00".to_string(),
            "#aa33aa".to_string(),
        ])
        .build()
        .unwrap()
}
